@@ -1,11 +1,18 @@
-import logging
+# encoding: utf-8
+
+from __future__ import absolute_import, division, print_function, unicode_literals
+
 import operator
-import re
 import warnings
-from django.conf import settings
+
+from django.utils import six
+
+from haystack import connection_router, connections
 from haystack.backends import SQ
-from haystack.constants import REPR_OUTPUT_SIZE, ITERATOR_LOAD_PER_QUERY, DEFAULT_OPERATOR
-from haystack.exceptions import NotRegistered
+from haystack.constants import DEFAULT_OPERATOR, ITERATOR_LOAD_PER_QUERY, REPR_OUTPUT_SIZE
+from haystack.exceptions import NotHandled
+from haystack.inputs import AutoQuery, Raw
+from haystack.utils import log as logging
 
 
 class SearchQuerySet(object):
@@ -14,12 +21,17 @@ class SearchQuerySet(object):
 
     Supports chaining (a la QuerySet) to narrow the search.
     """
-    def __init__(self, site=None, query=None):
+    def __init__(self, using=None, query=None):
+        # ``_using`` should only ever be a value other than ``None`` if it's
+        # been forced with the ``.using`` method.
+        self._using = using
+        self.query = None
+        self._determine_backend()
+
+        # If ``query`` is present, it should override even what the routers
+        # think.
         if query is not None:
             self.query = query
-        else:
-            from haystack import backend
-            self.query = backend.SearchQuery(site=site)
 
         self._result_cache = []
         self._result_count = None
@@ -28,11 +40,31 @@ class SearchQuerySet(object):
         self._ignored_result_count = 0
         self.log = logging.getLogger('haystack')
 
-        if site is not None:
-            self.site = site
+    def _determine_backend(self):
+        from haystack import connections
+        # A backend has been manually selected. Use it instead.
+        if self._using is not None:
+            self.query = connections[self._using].get_query()
+            return
+
+        # No backend, so rely on the routers to figure out what's right.
+        hints = {}
+
+        if self.query:
+            hints['models'] = self.query.models
+
+        backend_alias = connection_router.for_read(**hints)
+
+        if isinstance(backend_alias, (list, tuple)) and len(backend_alias):
+            # We can only effectively read from one engine.
+            backend_alias = backend_alias[0]
+
+        # The ``SearchQuery`` might swap itself out for a different variant
+        # here.
+        if self.query:
+            self.query = self.query.using(backend_alias)
         else:
-            from haystack import site as main_site
-            self.site = main_site
+            self.query = connections[backend_alias].get_query()
 
     def __getstate__(self):
         """
@@ -41,17 +73,14 @@ class SearchQuerySet(object):
         len(self)
         obj_dict = self.__dict__.copy()
         obj_dict['_iter'] = None
-        del obj_dict['site']
         obj_dict['log'] = None
         return obj_dict
 
-    def __setstate__(self, dict):
+    def __setstate__(self, data_dict):
         """
         For unpickling.
         """
-        self.__dict__ = dict
-        from haystack import site as main_site
-        self.site = main_site
+        self.__dict__ = data_dict
         self.log = logging.getLogger('haystack')
 
     def __repr__(self):
@@ -135,13 +164,13 @@ class SearchQuerySet(object):
             if not self._fill_cache(current_position, current_position + ITERATOR_LOAD_PER_QUERY):
                 raise StopIteration
 
-    def _fill_cache(self, start, end):
+    def _fill_cache(self, start, end, **kwargs):
         # Tell the query where to start from and how many we'd like.
         self.query._reset()
         self.query.set_limits(start, end)
-        results = self.query.get_results()
+        results = self.query.get_results(**kwargs)
 
-        if results == None or len(results) == 0:
+        if results is None or len(results) == 0:
             return False
 
         # Setup the full cache now that we know how many results there are.
@@ -151,7 +180,7 @@ class SearchQuerySet(object):
         # an array of 100,000 ``None``s consumed less than .5 Mb, which ought
         # to be an acceptable loss for consistent and more efficient caching.
         if len(self._result_cache) == 0:
-            self._result_cache = [None for i in xrange(self.query.get_count())]
+            self._result_cache = [None] * self.query.get_count()
 
         if start is None:
             start = 0
@@ -159,33 +188,41 @@ class SearchQuerySet(object):
         if end is None:
             end = self.query.get_count()
 
+        to_cache = self.post_process_results(results)
+
+        # Assign by slice.
+        self._result_cache[start:start + len(to_cache)] = to_cache
+        return True
+
+    def post_process_results(self, results):
+        to_cache = []
+
         # Check if we wish to load all objects.
         if self._load_all:
-            original_results = []
             models_pks = {}
             loaded_objects = {}
 
             # Remember the search position for each result so we don't have to resort later.
             for result in results:
-                original_results.append(result)
                 models_pks.setdefault(result.model, []).append(result.pk)
 
             # Load the objects for each model in turn.
             for model in models_pks:
                 try:
-                    loaded_objects[model] = self.site.get_index(model).read_queryset().in_bulk(models_pks[model])
-                except NotRegistered:
-                    self.log.warning("Model not registered with search site '%s.%s'." % (self.app_label, self.model_name))
+                    ui = connections[self.query._using].get_unified_index()
+                    index = ui.get_index(model)
+                    objects = index.read_queryset(using=self.query._using)
+                    loaded_objects[model] = objects.in_bulk(models_pks[model])
+                except NotHandled:
+                    self.log.warning("Model '%s' not handled by the routers", model)
                     # Revert to old behaviour
                     loaded_objects[model] = model._default_manager.in_bulk(models_pks[model])
 
-        to_cache = []
-
         for result in results:
             if self._load_all:
                 # We have to deal with integer keys being cast from strings
                 model_objects = loaded_objects.get(result.model, {})
-                if not result.pk in model_objects:
+                if result.pk not in model_objects:
                     try:
                         result.pk = int(result.pk)
                     except ValueError:
@@ -200,16 +237,13 @@ class SearchQuerySet(object):
 
             to_cache.append(result)
 
-        # Assign by slice.
-        self._result_cache[start:start + len(to_cache)] = to_cache
-        return True
-
+        return to_cache
 
     def __getitem__(self, k):
         """
         Retrieves an item or slice from the set of results.
         """
-        if not isinstance(k, (slice, int, long)):
+        if not isinstance(k, (slice, six.integer_types)):
             raise TypeError
         assert ((not isinstance(k, slice) and (k >= 0))
                 or (isinstance(k, slice) and (k.start is None or k.start >= 0)
@@ -232,7 +266,8 @@ class SearchQuerySet(object):
             bound = k + 1
 
         # We need check to see if we need to populate more of the cache.
-        if len(self._result_cache) <= 0 or (None in self._result_cache[start:bound] and not self._cache_is_full()):
+        if len(self._result_cache) <= 0 or (None in self._result_cache[start:bound]
+                                            and not self._cache_is_full()):
             try:
                 self._fill_cache(start, bound)
             except StopIteration:
@@ -245,15 +280,13 @@ class SearchQuerySet(object):
         else:
             return self._result_cache[start]
 
-
     # Methods that return a SearchQuerySet.
-
     def all(self):
         """Returns all results for the query."""
         return self._clone()
 
     def none(self):
-        """Returns all results for the query."""
+        """Returns an empty result list for the query."""
         return self._clone(klass=EmptySearchQuerySet)
 
     def filter(self, *args, **kwargs):
@@ -301,8 +334,8 @@ class SearchQuerySet(object):
         clone = self._clone()
 
         for model in models:
-            if not model in self.site.get_indexed_models():
-                warnings.warn('The model %r is not registered for search.' % model)
+            if model not in connections[self.query._using].get_unified_index().get_indexed_models():
+                warnings.warn('The model %r is not registered for search.' % (model,))
 
             clone.query.add_model(model)
 
@@ -325,10 +358,48 @@ class SearchQuerySet(object):
         clone.query.add_boost(term, boost)
         return clone
 
-    def facet(self, field):
+    def facet(self, field, **options):
         """Adds faceting to a query for the provided field."""
         clone = self._clone()
-        clone.query.add_field_facet(field)
+        clone.query.add_field_facet(field, **options)
+        return clone
+
+    def within(self, field, point_1, point_2):
+        """Spatial: Adds a bounding box search to the query."""
+        clone = self._clone()
+        clone.query.add_within(field, point_1, point_2)
+        return clone
+
+    def dwithin(self, field, point, distance):
+        """Spatial: Adds a distance-based search to the query."""
+        clone = self._clone()
+        clone.query.add_dwithin(field, point, distance)
+        return clone
+
+    def stats(self, field):
+        """Adds stats to a query for the provided field."""
+        return self.stats_facet(field, facet_fields=None)
+
+    def stats_facet(self, field, facet_fields=None):
+        """Adds stats facet for the given field and facet_fields represents
+        the faceted fields."""
+        clone = self._clone()
+        stats_facets = []
+        try:
+            stats_facets.append(sum(facet_fields, []))
+        except TypeError:
+            if facet_fields:
+                stats_facets.append(facet_fields)
+        clone.query.add_stats_query(field, stats_facets)
+        return clone
+
+    def distance(self, field, point):
+        """
+        Spatial: Denotes results must have distance measurements from the
+        provided point.
+        """
+        clone = self._clone()
+        clone.query.add_distance(field, point)
         return clone
 
     def date_facet(self, field, start_date, end_date, gap_by, gap_amount=1):
@@ -345,15 +416,20 @@ class SearchQuerySet(object):
 
     def narrow(self, query):
         """Pushes existing facet choices into the search."""
+
+        if isinstance(query, SQ):
+            # produce query string using empty query of the same class
+            empty_query = self.query._clone()
+            empty_query._reset()
+            query = query.as_query_string(empty_query.build_query_fragment)
+
         clone = self._clone()
         clone.query.add_narrow_query(query)
         return clone
 
     def raw_search(self, query_string, **kwargs):
         """Passes a raw query directly to the backend."""
-        clone = self._clone()
-        clone.query.raw_search(query_string, **kwargs)
-        return clone
+        return self.filter(content=Raw(query_string, **kwargs))
 
     def load_all(self):
         """Efficiently populates the objects in the search results."""
@@ -361,51 +437,17 @@ class SearchQuerySet(object):
         clone._load_all = True
         return clone
 
-    def auto_query(self, query_string):
+    def auto_query(self, query_string, fieldname='content'):
         """
         Performs a best guess constructing the search query.
 
         This method is somewhat naive but works well enough for the simple,
         common cases.
         """
-        clone = self._clone()
-
-        # Pull out anything wrapped in quotes and do an exact match on it.
-        open_quote_position = None
-        non_exact_query = query_string
-
-        for offset, char in enumerate(query_string):
-            if char == '"':
-                if open_quote_position != None:
-                    current_match = non_exact_query[open_quote_position + 1:offset]
-
-                    if current_match:
-                        clone = clone.filter(content=clone.query.clean(current_match))
-
-                    non_exact_query = non_exact_query.replace('"%s"' % current_match, '', 1)
-                    open_quote_position = None
-                else:
-                    open_quote_position = offset
-
-        # Pseudo-tokenize the rest of the query.
-        keywords = non_exact_query.split()
-
-        # Loop through keywords and add filters to the query.
-        for keyword in keywords:
-            exclude = False
-
-            if keyword.startswith('-') and len(keyword) > 1:
-                keyword = keyword[1:]
-                exclude = True
-
-            cleaned_keyword = clone.query.clean(keyword)
-
-            if exclude:
-                clone = clone.exclude(content=cleaned_keyword)
-            else:
-                clone = clone.filter(content=cleaned_keyword)
-
-        return clone
+        kwargs = {
+            fieldname: AutoQuery(query_string)
+        }
+        return self.filter(**kwargs)
 
     def autocomplete(self, **kwargs):
         """
@@ -420,12 +462,23 @@ class SearchQuerySet(object):
         for field_name, query in kwargs.items():
             for word in query.split(' '):
                 bit = clone.query.clean(word.strip())
-                kwargs = {
-                    field_name: bit,
-                }
-                query_bits.append(SQ(**kwargs))
+                if bit:
+                    kwargs = {
+                        field_name: bit,
+                    }
+                    query_bits.append(SQ(**kwargs))
 
-        return clone.filter(reduce(operator.__and__, query_bits))
+        return clone.filter(six.moves.reduce(operator.__and__, query_bits))
+
+    def using(self, connection_name):
+        """
+        Allows switching which connection the ``SearchQuerySet`` uses to
+        search in.
+        """
+        clone = self._clone()
+        clone.query = self.query.using(connection_name)
+        clone._using = connection_name
+        return clone
 
     # Methods that do not return a SearchQuerySet.
 
@@ -457,22 +510,65 @@ class SearchQuerySet(object):
         This will cause the query to execute and should generally be used when
         presenting the data.
         """
-        clone = self._clone()
-        return clone.query.get_facet_counts()
+        if self.query.has_run():
+            return self.query.get_facet_counts()
+        else:
+            clone = self._clone()
+            return clone.query.get_facet_counts()
+
+    def stats_results(self):
+        """
+        Returns the stats results found by the query.
+        """
+        if self.query.has_run():
+            return self.query.get_stats()
+        else:
+            clone = self._clone()
+            return clone.query.get_stats()
 
     def spelling_suggestion(self, preferred_query=None):
         """
         Returns the spelling suggestion found by the query.
 
-        To work, you must set ``settings.HAYSTACK_INCLUDE_SPELLING`` to True.
-        Otherwise, ``None`` will be returned.
+        To work, you must set ``INCLUDE_SPELLING`` within your connection's
+        settings dictionary to ``True``. Otherwise, ``None`` will be returned.
 
         This will cause the query to execute and should generally be used when
         presenting the data.
         """
-        clone = self._clone()
-        return clone.query.get_spelling_suggestion(preferred_query)
+        if self.query.has_run():
+            return self.query.get_spelling_suggestion(preferred_query)
+        else:
+            clone = self._clone()
+            return clone.query.get_spelling_suggestion(preferred_query)
 
+    def values(self, *fields):
+        """
+        Returns a list of dictionaries, each containing the key/value pairs for
+        the result, exactly like Django's ``ValuesQuerySet``.
+        """
+        qs = self._clone(klass=ValuesSearchQuerySet)
+        qs._fields.extend(fields)
+        return qs
+
+    def values_list(self, *fields, **kwargs):
+        """
+        Returns a list of field values as tuples, exactly like Django's
+        ``QuerySet.values``.
+
+        Optionally accepts a ``flat=True`` kwarg, which in the case of a
+        single field being provided, will return a flat list of that field
+        rather than a list of tuples.
+        """
+        flat = kwargs.pop("flat", False)
+
+        if flat and len(fields) > 1:
+            raise TypeError("'flat' is not valid when values_list is called with more than one field.")
+
+        qs = self._clone(klass=ValuesListSearchQuerySet)
+        qs._fields.extend(fields)
+        qs._flat = flat
+        return qs
 
     # Utility methods.
 
@@ -481,7 +577,7 @@ class SearchQuerySet(object):
             klass = self.__class__
 
         query = self.query._clone()
-        clone = klass(site=self.site, query=query)
+        clone = klass(query=query)
         clone._load_all = self._load_all
         return clone
 
@@ -510,6 +606,72 @@ class EmptySearchQuerySet(SearchQuerySet):
         return {}
 
 
+class ValuesListSearchQuerySet(SearchQuerySet):
+    """
+    A ``SearchQuerySet`` which returns a list of field values as tuples, exactly
+    like Django's ``ValuesListQuerySet``.
+    """
+    def __init__(self, *args, **kwargs):
+        super(ValuesListSearchQuerySet, self).__init__(*args, **kwargs)
+        self._flat = False
+        self._fields = []
+
+        # Removing this dependency would require refactoring much of the backend
+        # code (_process_results, etc.) and these aren't large enough to make it
+        # an immediate priority:
+        self._internal_fields = ['id', 'django_ct', 'django_id', 'score']
+
+    def _clone(self, klass=None):
+        clone = super(ValuesListSearchQuerySet, self)._clone(klass=klass)
+        clone._fields = self._fields
+        clone._flat = self._flat
+        return clone
+
+    def _fill_cache(self, start, end):
+        query_fields = set(self._internal_fields)
+        query_fields.update(self._fields)
+        kwargs = {
+            'fields': query_fields
+        }
+        return super(ValuesListSearchQuerySet, self)._fill_cache(start, end, **kwargs)
+
+    def post_process_results(self, results):
+        to_cache = []
+
+        if self._flat:
+            accum = to_cache.extend
+        else:
+            accum = to_cache.append
+
+        for result in results:
+            accum([getattr(result, i, None) for i in self._fields])
+
+        return to_cache
+
+
+class ValuesSearchQuerySet(ValuesListSearchQuerySet):
+    """
+    A ``SearchQuerySet`` which returns a list of dictionaries, each containing
+    the key/value pairs for the result, exactly like Django's
+    ``ValuesQuerySet``.
+    """
+    def _fill_cache(self, start, end):
+        query_fields = set(self._internal_fields)
+        query_fields.update(self._fields)
+        kwargs = {
+            'fields': query_fields
+        }
+        return super(ValuesListSearchQuerySet, self)._fill_cache(start, end, **kwargs)
+
+    def post_process_results(self, results):
+        to_cache = []
+
+        for result in results:
+            to_cache.append(dict((i, getattr(result, i, None)) for i in self._fields))
+
+        return to_cache
+
+
 class RelatedSearchQuerySet(SearchQuerySet):
     """
     A variant of the SearchQuerySet that can handle `load_all_queryset`s.
@@ -518,8 +680,11 @@ class RelatedSearchQuerySet(SearchQuerySet):
     far less efficient but needs to fill the cache before it to maintain
     consistency.
     """
-    _load_all_querysets = {}
-    _result_cache = []
+
+    def __init__(self, *args, **kwargs):
+        super(RelatedSearchQuerySet, self).__init__(*args, **kwargs)
+        self._load_all_querysets = {}
+        self._result_cache = []
 
     def _cache_is_full(self):
         return len(self._result_cache) >= len(self)
@@ -566,13 +731,11 @@ class RelatedSearchQuerySet(SearchQuerySet):
 
         # Check if we wish to load all objects.
         if self._load_all:
-            original_results = []
             models_pks = {}
             loaded_objects = {}
 
             # Remember the search position for each result so we don't have to resort later.
             for result in results:
-                original_results.append(result)
                 models_pks.setdefault(result.model, []).append(result.pk)
 
             # Load the objects for each model in turn.
@@ -583,12 +746,12 @@ class RelatedSearchQuerySet(SearchQuerySet):
                 else:
                     # Check the SearchIndex for the model for an override.
                     try:
-                        index = self.site.get_index(model)
+                        index = connections[self.query._using].get_unified_index().get_index(model)
                         qs = index.load_all_queryset()
                         loaded_objects[model] = qs.in_bulk(models_pks[model])
-                    except NotRegistered:
-                        # The model returned doesn't seem to be registered with
-                        # the current site. We should silently fail and populate
+                    except NotHandled:
+                        # The model returned doesn't seem to be handled by the
+                        # routers. We should silently fail and populate
                         # nothing for those objects.
                         loaded_objects[model] = []
 
@@ -619,8 +782,9 @@ class RelatedSearchQuerySet(SearchQuerySet):
         """
         Retrieves an item or slice from the set of results.
         """
-        if not isinstance(k, (slice, int, long)):
+        if not isinstance(k, (slice, six.integer_types)):
             raise TypeError
+
         assert ((not isinstance(k, slice) and (k >= 0))
                 or (isinstance(k, slice) and (k.start is None or k.start >= 0)
                     and (k.stop is None or k.stop >= 0))), \
@@ -674,7 +838,7 @@ class RelatedSearchQuerySet(SearchQuerySet):
             klass = self.__class__
 
         query = self.query._clone()
-        clone = klass(site=self.site, query=query)
+        clone = klass(query=query)
         clone._load_all = self._load_all
         clone._load_all_querysets = self._load_all_querysets
         return clone
